from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)


class ScopeBasedPolicy(BaseAuthorizationPolicy):
    """
    An authorization policy that allows access if you have the correct scopes.
    """

    def authorize(self, policy_context, user_context, request_context):
        for resource_scope in request_context.resource_scopes:
            if resource_scope in user_context.scopes:
                policy_context.access_verdict = True

        return policy_context



import json
import os
import requests

from . import flask_process, custom_token
from dotenv import load_dotenv


load_dotenv()


class TestPolicyFactory:
    ENDPOINT = str(os.getenv("ENDPOINT"))
    SUPER_ADMIN_ROLE = "super_admin"
    REGULAR_USER_ROLE = "regular_user"

    @classmethod
    def setup_class(cls):
        flask_process.set_app_policies(
            ["token_based_policies.authlib_flask_oauth2_policy"],
            ["super_admin_policy", "scope_based_policy", "open_data_policy"],
        )
        flask_process.start()

    def setup_method(self):
        flask_process.assert_running()

    def test_request_without_token_returns_401(self):
        response = requests.put(self.ENDPOINT)
        assert response.status_code == 401

    def test_request_with_wrong_token_payload_structure_returns_403(self):
        payload = {
            "azp": "inuits-policy-based-auth",
            "inuits-policy-based-auth": {"roles": [self.SUPER_ADMIN_ROLE]},
        }
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_with_wrong_token_azp_returns_403(self):
        payload = {
            "azp": "inuits-policy-based-auth",
            "resource_access": {
                "policy-based-auth": {"roles": [self.SUPER_ADMIN_ROLE]}
            },
        }
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_with_token_returns_200(self):
        payload = self._get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.put(self.ENDPOINT, headers=headers)

        assert response.status_code == 200

    def test_super_admin_gets_correct_user_context(self):
        payload = self._get_payload([self.SUPER_ADMIN_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        assert response.status_code == 200
        assert response.json()["roles"] == [self.SUPER_ADMIN_ROLE]

    def test_regular_user_with_invalid_scopes_cannot_successfully_do_post_request(self):
        payload = self._get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_regular_user_can_successfully_do_get_request(self):
        payload = self._get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        json_response = response.json()
        assert response.status_code == 200
        assert json_response["roles"] == [self.REGULAR_USER_ROLE]
        assert json_response["scopes"] == self._get_scopes(self.REGULAR_USER_ROLE)

    def _get_payload(self, roles):
        return {
            "azp": "inuits-policy-based-auth",
            "resource_access": {"inuits-policy-based-auth": {"roles": roles}},
        }

    def _get_scopes(self, role):
        with open(str(os.getenv("TEST_API_SCOPES")), "r") as scopes_file:
            return json.load(scopes_file)[role]

    @classmethod
    def teardown_class(cls):
        flask_process.stop()



from collections.abc import MutableMapping
from inuits_policy_based_auth.helpers.access_restrictions import AccessRestrictions
from inuits_policy_based_auth.helpers.immutable_dict import ImmutableDict


class UserContext:
    """
    A class containing data about the authenticated user.

    Properties
    ----------
    auth_objects : ImmutableDict
        An immutable dict containing objects used to authenticate a user,
        for example a token.
    email : str
        The email of the authenticated user.
    roles : list[str]
        The roles of the authenticated user.
    scopes : list[str]
        The scopes of the authenticated user.
    tenant_names : list[str]
        Names of the tenants the authenticated user is part of.
    tenant_objects : list[str]
        Object representations of the tenants the authenticated user is part of.
    access_restrictions : AccessRestrictions
        A class containing properties used to restrict access.

    Methods
    -------
    flatten_auth_object(data, parent_key)
        Flattens the auth object to be a dict of one level deep.
    """

    def __init__(self):
        self._auth_objects = ImmutableDict({})
        self._email = ""
        self._roles = []
        self._scopes = []
        self._tenant_names = []
        self._tenant_objects = []
        self._access_restrictions = AccessRestrictions()

    @property
    def auth_objects(self):
        """
        An immutable dict containing objects used to authenticate a user,
        for example a token.
        """

        return self._auth_objects

    @property
    def email(self):
        """The email of the authenticated user."""

        return self._email

    @email.setter
    def email(self, email: str):
        self._email = email

    @property
    def roles(self):
        """The roles of the authenticated user."""

        return self._roles

    @roles.setter
    def roles(self, roles: list[str]):
        self._roles = roles

    @property
    def scopes(self):
        """The scopes of the authenticated user."""

        return self._scopes

    @scopes.setter
    def scopes(self, scopes: list[str]):
        self._scopes = scopes

    @property
    def tenant_names(self):
        """Names of the tenants the authenticated user is part of."""

        return self._tenant_names

    @tenant_names.setter
    def tenant_names(self, names: list[str]):
        self._tenant_names = names

    @property
    def tenant_objects(self):
        """
        Object representations of the tenants the authenticated user is part of.
        """

        return self._tenant_objects

    @tenant_objects.setter
    def tenant_objects(self, objects: list):
        self._tenant_objects = objects

    @property
    def access_restrictions(self):
        """A class containing properties used to restrict access."""

        return self._access_restrictions

    def __flatten_auth_object_generator(self, data: MutableMapping, parent_key):
        for key, value in data.items():
            flattened_key = parent_key + "." + key if parent_key else key
            if isinstance(value, MutableMapping):
                yield from self.flatten_auth_object(value, flattened_key).items()
            else:
                yield flattened_key, value

    def flatten_auth_object(self, data: MutableMapping, parent_key=""):
        """Flattens the auth object to be a dict of one level deep.

        Parameters
        ----------
        data : MutableMapping
            An object used to authenticate a user with, for example a token.
        parent_key : str, optional
            A key that will be the root key for the flattened dict.

        Returns
        -------
        dict
            A flattened dictionary representation of the auth object.
        """

        return dict(self.__flatten_auth_object_generator(data, parent_key))



from .policy_loader import load_policies
from flask import Flask, make_response
from flask_restful import Api, Resource, request
from inuits_policy_based_auth import PolicyFactory, RequestContext
from logging import Logger


app = Flask(__name__)
api = Api(app)

policy_factory = PolicyFactory()
load_policies(policy_factory, Logger(""))


class Entity(Resource):
    @policy_factory.apply_policies(
        RequestContext(request, ["read-entity", "update-entity"])
    )
    def get(self):
        user_context = policy_factory.get_user_context()
        response_body = {
            "auth_objects": user_context.auth_objects.get("token"),
            "email": user_context.email,
            "roles": user_context.roles,
            "scopes": user_context.scopes,
            "tenant_names": user_context.tenant_names,
            "tenant_objects": user_context.tenant_objects,
        }
        return make_response(response_body, 200)

    @policy_factory.apply_policies(RequestContext(request))
    def post(self):
        user_context = policy_factory.get_user_context()
        response_body = {
            "auth_objects": user_context.auth_objects.get("token"),
            "email": user_context.email,
            "roles": user_context.roles,
            "scopes": user_context.scopes,
            "tenant_names": user_context.tenant_names,
            "tenant_objects": user_context.tenant_objects,
        }
        return make_response(response_body, 201)

    @policy_factory.authenticate()
    def put(self):
        user_context = policy_factory.get_user_context()
        response_body = {
            "auth_objects": user_context.auth_objects.get("token"),
            "email": user_context.email,
            "roles": user_context.roles,
            "scopes": user_context.scopes,
            "tenant_names": user_context.tenant_names,
            "tenant_objects": user_context.tenant_objects,
        }
        return make_response(response_body, 200)


api.add_resource(Entity, "/")



import jwt
import os

from datetime import datetime, timezone, timedelta


def get_authorization_header(testcase_specific_payload, token_algorithm="RS256"):
    payload = {
        "exp": datetime.now(tz=timezone.utc) + timedelta(seconds=10),
        "iss": str(os.getenv("STATIC_ISSUER")),
        "sub": str(os.getenv("STATIC_ISSUER")),
    }

    private_key = f"-----BEGIN PRIVATE KEY-----\n{os.getenv('STATIC_PRIVATE_KEY')}\n-----END PRIVATE KEY-----"
    testcase_specific_payload.update(payload)

    token = jwt.encode(
        testcase_specific_payload, private_key, algorithm=token_algorithm
    )
    return {"Authorization": f"Bearer {token}"}



import json
import os
import pytest

from subprocess import Popen, DEVNULL
from time import sleep


_flask_process: Popen[bytes]
_sleep_delay = 1


def set_app_policies(authentication: list[str], authorization: list[str]):
    _overwrite_configuration_file(authentication, authorization)


def start():
    global _flask_process
    _flask_process = Popen(["flask", "run"], stdout=DEVNULL, stderr=DEVNULL)
    sleep(_sleep_delay)


def assert_running():
    if _flask_process.poll() != None:
        pytest.fail("flask_process is not running.")


def stop():
    _overwrite_configuration_file()
    _flask_process.kill()
    sleep(_sleep_delay)
    assert _flask_process.poll() != None


def _overwrite_configuration_file(
    authentication: list[str] = [], authorization: list[str] = []
):
    with open(str(os.getenv("TEST_API_CONFIGURATION")), "r+") as configuration_file:
        configuration = json.load(configuration_file)

        for app in ["test_api", "test_api_backup"]:
            if authentication and authorization:
                configuration[app]["policies"]["authentication"].extend(authentication)
                configuration[app]["policies"]["authorization"].extend(authorization)
            else:
                configuration[app]["policies"]["authentication"] = []
                configuration[app]["policies"]["authorization"] = []

        configuration_file.seek(0)
        configuration_file.write(json.dumps(configuration, indent=2))
        configuration_file.truncate()



class PolicyFactoryException(Exception):
    def __init__(self, message):
        super().__init__(message)


class NoUserContextException(PolicyFactoryException):
    def __init__(self):
        super().__init__(
            "Cannot get an object of type UserContext if no policies are applied yet."
        )


class NoAuthenticationPoliciesToApplyException(PolicyFactoryException):
    def __init__(self):
        super().__init__("No authentication policies set to apply.")


class NoAuthorizationPoliciesToApplyException(PolicyFactoryException):
    def __init__(self):
        super().__init__("No authorization policies set to apply.")


class InvalidFallbackKey(PolicyFactoryException):
    def __init__(self, key: str):
        super().__init__(
            f"Provided fallback key '{key}' is not registered in either authentication policies, authorization policies, or both."
        )


class NoFallbackKeySet(PolicyFactoryException):
    def __init__(self):
        super().__init__("No fallback key for policy mapping is set.")


class PolicyException(Exception):
    pass


class AuthenticateMethodDidNotReturnObjectOfTypeUserContextException(PolicyException):
    def __init__(self):
        super().__init__(
            "Authenticate method of authentication policy did not return an object of type UserContext."
        )


class AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException(PolicyException):
    def __init__(self):
        super().__init__(
            "Authorize method of authorization policy did not return an object of type PolicyContext."
        )



import json
import os

from importlib import import_module
from inuits_policy_based_auth import PolicyFactory
from inuits_policy_based_auth.exceptions import (
    PolicyFactoryException,
)
from logging import Logger


def load_policies(policy_factory: PolicyFactory, logger: Logger):
    apps = {}

    with open(str(os.getenv("TEST_API_CONFIGURATION")), "r") as configuration_file:
        apps = json.load(configuration_file)

    for app in apps:
        try:
            auth_type = "authentication"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy = __instantiate_authentication_policy(
                    policy_module_name, policy, logger
                )
                policy_factory.register_authentication_policy(
                    f"tests.integration.{app}", policy
                )

            auth_type = "authorization"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy_factory.register_authorization_policy(
                    f"tests.integration.{app}", policy()
                )
        except Exception as error:
            raise PolicyFactoryException(
                f"Policy factory was not configured correctly: {str(error)}"
            ).with_traceback(error.__traceback__)

    policy_factory.set_fallback_key_for_policy_mapping("tests.integration.test_api")


def __get_class(app, auth_type, policy_module_name):
    module = None

    try:
        module = import_module(f"apps.{app}.policies.{auth_type}.{policy_module_name}")
    except:
        module = import_module(
            f"inuits_policy_based_auth.{auth_type}.policies.{policy_module_name}"
        )

    policy_class_name = module.__name__.split(".")[-1].title().replace("_", "")
    policy = getattr(module, policy_class_name)
    return policy


def __instantiate_authentication_policy(policy_module_name, policy, logger: Logger):
    if policy_module_name == "token_based_policies.authlib_flask_oauth2_policy":
        return policy(
            logger,
            os.getenv("ROLE_SCOPE_MAPPING", os.getenv("TEST_API_SCOPES")),
            os.getenv("STATIC_ISSUER"),
            os.getenv("STATIC_PUBLIC_KEY"),
        )

    return policy()



from inuits_policy_based_auth.authorization.policies.scope_based_policy import (
    ScopeBasedPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext


class TestScopeBasedPolicy:
    def setup_method(self):
        self.scope_based_policy = ScopeBasedPolicy()
        self.policy_context = PolicyContext()
        self.user_context = UserContext()

    def test_authorize_allows_access(self):
        self.user_context.scopes = ["read"]
        request_context = RequestContext(None, ["create", "read"])

        policy_context = self.scope_based_policy.authorize(
            self.policy_context, self.user_context, request_context
        )

        assert policy_context.access_verdict == True

    def test_authorize_does_not_determine_access(self):
        self.user_context.scopes = ["update"]
        request_context = RequestContext(None, ["create", "read"])

        policy_context = self.scope_based_policy.authorize(
            self.policy_context, self.user_context, request_context
        )

        assert policy_context.access_verdict == None
